import numpy as np
from mne import Epochs
from mne.preprocessing import bads, ICA
from autoreject import autoreject, Ransac
from random import sample

from .utils.config import settings
from mne.utils import logger


def prepare_epochs_for_ica(epochs: Epochs) -> Epochs:
    """
    Drops epochs that were marked bad based on a global outlier detection.
    This implementation for the preliminary epoch rejection was based on the
    Python implementation of the FASTER algorithm from Marijn van Vliet
    https://gist.github.com/wmvanvliet/d883c3fe1402c7ced6fc
    Parameters
    ----------
    epochs

    Returns
    -------
    Epochs instance
    """
    logger.info('Preliminary epoch rejection: ')

    def _deviation(data: np.ndarray) -> np.ndarray:
        """
        Computes the deviation from mean for each channel.
        """
        channels_mean = np.mean(data, axis=2)
        return channels_mean - np.mean(channels_mean, axis=0)

    metrics = {
        'amplitude': lambda x: np.mean(np.ptp(x, axis=2), axis=1),
        'deviation': lambda x: np.mean(_deviation(x), axis=1),
        'variance': lambda x: np.mean(np.var(x, axis=2), axis=1),
    }

    epochs_data = epochs.get_data()

    bad_epochs = []
    for metric in metrics:
        scores = metrics[metric](epochs_data)
        outliers = bads._find_outliers(scores, threshold=3.0)
        logger.info(f'Bad epochs by {metric}\n\t{outliers}')
        bad_epochs.extend(outliers)

    bad_epochs = list(set(bad_epochs))
    epochs_faster = epochs.copy().drop(bad_epochs, reason='FASTER')

    return epochs_faster


def run_ica(epochs: Epochs) -> ICA:
    """
    Runs ICA decomposition on Epochs instance.

    If there are no EOG channels found, it tries to use 'Fp1' and 'Fp2' as EOG
    channels; if they are not found either, it chooses the first two channels
    to identify EOG components with mne.preprocessing.ica.find_bads_eog().
    Parameters
    ----------
    epochs: the instance to be used for ICA decomposition
    Returns
    -------
    ICA instance
    """
    ica = ICA(n_components=settings['ica']['n_components'],
              random_state=42,
              method=settings['ica']['method'])
    ica_epochs = epochs.copy()
    ica.fit(ica_epochs, decim=settings['ica']['decim'])

    if 'eog' not in epochs.get_channel_types():
        if 'Fp1' and 'Fp2' in epochs.get_montage().ch_names:
            eog_channels = ['Fp1', 'Fp2']
        else:
            eog_channels = epochs.get_montage().ch_names[:2]
        logger.info('EOG channels are not found. Attempting to use '
                    f'{",".join(eog_channels)} channels as EOG channels.')
        ica_epochs.set_channel_types({ch: 'eog' for ch in eog_channels})

    eog_indices, _ = ica.find_bads_eog(ica_epochs)
    ica.exclude = eog_indices

    return ica


def run_autoreject(epochs: Epochs, n_jobs: int = 11,
                   subset: bool = False) -> autoreject.RejectLog:
    """
    Drop bad epochs based on AutoReject.
    Parameters
    ----------
    epochs: the instance to be cleaned
    n_jobs: the number of parallel processes to be run
    subset: whether to train autoreject on a random subset of data (faster)

    Returns
    -------
    Autoreject instance
    """
    if not epochs.preload:
        epochs_autoreject = epochs.copy().load_data()
    else:
        epochs_autoreject = epochs.copy()

    ar = autoreject.AutoReject(random_state=42, n_jobs=n_jobs)

    n_epochs = len(epochs_autoreject)
    if subset:
        logger.info(f'Fitting autoreject on random (n={int(n_epochs * 0.25)}) '
                    f'subset of epochs: ')
        subset = sample(set(np.arange(0, n_epochs, 1)), int(n_epochs * 0.25))
        ar.fit(epochs_autoreject[subset])

    else:
        logger.info(f'Fitting autoreject on (n={n_epochs}) epochs: ')
        ar.fit(epochs_autoreject)

    reject_log = ar.get_reject_log(epochs_autoreject)

    # report bad epochs where more than 15% (value updated from config.py)
    # of channels were marked as noisy within an epoch
    threshold = settings['autoreject']['threshold']
    num_bad_epochs = np.count_nonzero(reject_log.labels, axis=1)
    bad_epochs = np.where(num_bad_epochs > threshold * epochs.info['nchan'])
    bad_epochs = bad_epochs[0].tolist()

    reject_log.report = bad_epochs

    logger.info('\nAUTOREJECT report\n'
                f'There are {len(epochs_autoreject[reject_log.bad_epochs])} '
                f'bad epochs found with Autoreject. '
                f'You can assess these epochs with reject_log.bad_epochs\n'
                f'\nThere are {len(bad_epochs)} bad epochs where more than '
                f'{int(threshold * 100)}% of the channels were noisy. '
                f'You can assess these epochs with reject_log.report')

    return reject_log


def run_ransac(epochs: Epochs, n_jobs: int = 11) -> Ransac:
    """
    Find bad channels with Ransac.
    Parameters
    ----------
    epochs: the instance where bad channels to be found
    n_jobs: the number of parallel processes to run

    Returns
    -------
    Ransac instance
    """
    if not epochs.preload:
        epochs_ransac = epochs.copy().load_data()
    else:
        epochs_ransac = epochs.copy()

    ransac = Ransac(verbose='progressbar', n_jobs=n_jobs).fit(epochs_ransac)

    threshold = settings['ransac']['threshold']
    num_bad_channels = np.count_nonzero(ransac.bad_log, axis=0) / len(
        epochs_ransac)
    bad_channel_ids = np.where(num_bad_channels >= threshold)[0].tolist()

    num_bad_channels = num_bad_channels[bad_channel_ids]
    report_dict = dict(zip([epochs.ch_names[ch] for ch in bad_channel_ids],
                           np.round(num_bad_channels, 2)))
    report_dict = {k: v for k, v in
                   sorted(report_dict.items(), key=lambda item: item[1],
                          reverse=True)}

    if bad_channel_ids:
        logger.info('\nRANSAC report\n'
                    f'There are {len(num_bad_channels)} channels that were '
                    f'noisy for > {int(threshold * 100)}% of the time:')
        logger.info(
            "\n".join("{}\t{}".format(k, v) for k, v in report_dict.items()))

    if ransac.bad_chs_:
        logger.info(f'\nRANSAC marked {", ".join(ransac.bad_chs_)} '
                    f'channels to be interpolated.'
                    f'You can assess these channels with ransac.bad_chs_')

    else:
        logger.info(
            f'\nRANSAC did not mark any channels to be interpolated.'
            f'You can still assess the reported channels with ransac.report')

    ransac.report = list(report_dict.keys())

    return ransac



import os
from datetime import datetime
import numpy as np
import pandas as pd
from mne import Epochs


def update_log(log_file_path: str, epochs: Epochs, notes: str) -> pd.DataFrame:
    """
    Documents the changes during preprocessing for an Epochs object.
    Custom description can be added with the notes argument.
    The file path has to contain the file extension too (.csv).

    Parameters
    ----------
    log_file_path
    epochs
    notes

    Returns
    ----------
    log data
    """
    fid = epochs.info['fid']
    dropped_epochs_marker = ['FASTER', 'USER', 'AUTOREJECT']
    n_bad_epochs = len([drop for drop in epochs.drop_log if
                        np.isin(dropped_epochs_marker, drop).any()])

    log = pd.DataFrame({'fid': [fid],
                        'highpass': [epochs.info['highpass']],
                        'lowpass': [epochs.info['lowpass']],
                        'n_components': [np.NaN],
                        'n_bad_epochs': [n_bad_epochs],
                        'n_total_epochs': [len(epochs)],
                        'drop_percentage': [round(epochs.drop_log_stats(), 2)],
                        'stimuli': [list(epochs.event_id.keys())],
                        't_min': [epochs.tmin],
                        't_max': [epochs.tmax],
                        'n_interpolated': [np.NaN],
                        'average_ref_applied': [
                            bool(epochs.info['custom_ref_applied'])],
                        'baseline': [
                            epochs.baseline if epochs.baseline else np.NaN],
                        'notes': [notes],
                        'date_of_update': [datetime.utcnow().isoformat()]})

    description = epochs.info['description']
    if 'n_components' in description:
        n_components = [x for x in description.split(',')[0] if x.isdigit()][0]
        log['n_components'].update(int(n_components))

    if 'interpolated' in description:
        description_plain = description.replace(' ', '').replace(':', '')
        interpolated_channels_str = description_plain.split('interpolated')[1]
        n_interpolated = len(interpolated_channels_str.split(','))
        log['n_interpolated'].update(n_interpolated)

    if os.path.isfile(log_file_path):
        log.to_csv(log_file_path, mode='a', index=False, header=False)
    else:
        log.to_csv(log_file_path, index=False)

    return log



import os
import unittest
from mne import datasets


class TestRaw(unittest.TestCase):
    def setUp(self) -> None:
        self.raw_file_path = os.path.join(datasets.sample.data_path(), 'MEG',
                                          'sample', 'sample_audvis_raw.fif')

    def test_read_raw_measurement(self) -> None:
        from meeg_tools.utils.raw import read_raw_measurement

        raw = read_raw_measurement(raw_file_path=self.raw_file_path)
        assert raw.info['fid'] == 'sample_audvis_raw'
        assert bool(raw.get_montage())

    def test_filter_raw(self) -> None:
        from mne.io import read_raw
        from meeg_tools.utils.raw import filter_raw
        from meeg_tools.utils.config import settings

        settings['bandpass_filter']['low_freq'] = 1
        settings['bandpass_filter']['high_freq'] = 30

        raw = read_raw(fname=self.raw_file_path)

        raw_filtered = filter_raw(raw)

        assert raw_filtered.info['highpass'] == 1.0
        assert raw_filtered.info['lowpass'] == 30.0

        settings['bandpass_filter']['low_freq'] = 5
        settings['bandpass_filter']['high_freq'] = 45

        raw_filtered = filter_raw(raw)

        assert raw_filtered.info['highpass'] == 5.0
        assert raw_filtered.info['lowpass'] == 45.0

        assert raw.info['highpass'] != raw_filtered.info['highpass']
        assert raw.info['lowpass'] != raw_filtered.info['lowpass']



import os
import unittest
from mne import datasets
from mne.io import read_raw

from meeg_tools.utils.epochs import create_epochs


class TestPreprocessing(unittest.TestCase):
    def setUp(self) -> None:
        self.raw_file_path = os.path.join(datasets.sample.data_path(), 'MEG',
                                          'sample', 'sample_audvis_raw.fif')

        self.raw = read_raw(fname=self.raw_file_path)
        self.epochs = create_epochs(self.raw).load_data().pick_types(eeg=True)

    def test_prepare_epochs_for_ica(self):
        from meeg_tools.preprocessing import prepare_epochs_for_ica

        assert not any(self.epochs.drop_log)
        epochs_faster = prepare_epochs_for_ica(epochs=self.epochs)

        bad_epochs_indices = [drop_ind for drop_ind, drop in
                              enumerate(epochs_faster.drop_log) if drop]
        assert len(bad_epochs_indices) == 24
        assert bad_epochs_indices == [14, 30, 60, 100, 113, 123, 188, 218, 219,
                                      220, 221, 231, 235, 236, 237, 238, 242,
                                      243, 244, 267, 268, 269, 270, 271]

    def test_run_ica(self):
        from meeg_tools.preprocessing import run_ica
        from meeg_tools.utils.config import settings

        settings['ica']['n_components'] = 32
        settings['ica']['decim'] = 3
        ica = run_ica(epochs=self.epochs)

        assert ica.exclude == [0]
        assert ica.n_components == 32
        assert 'eog' not in self.epochs.get_channel_types()

    def test_run_autoreject(self):
        from meeg_tools.preprocessing import run_autoreject

        reject_log = run_autoreject(epochs=self.epochs, subset=False)
        epochs_autoreject = self.epochs.copy().drop(reject_log.report,
                                                    reason='AUTOREJECT')

        assert len(epochs_autoreject) < len(self.epochs)

    def test_run_ransac(self):
        from meeg_tools.preprocessing import run_ransac

        ransac = run_ransac(epochs=self.epochs)

        assert ransac.bad_chs_ == ['EEG 001', 'EEG 009']



import numpy as np
from matplotlib import pyplot as plt
from mne import Epochs
from mne.connectivity import spectral_connectivity
from mne.preprocessing import compute_current_source_density


def compute_connectivity(epochs: Epochs, **kwargs) -> np.ndarray:
    """
    Compute channel level connectivity matrix from Epochs instance.
    Returns the computed connectivity matrix (n_freqs, n_signals, n_signals).

    Args:
        str spectrum_mode: Valid estimation mode 'fourier' or 'multitaper'
        Epochs epochs: Epochs extracted from a Raw instance
        str method: connectivity estimation method
        int n_jobs: number of epochs to process in parallel
    :return: np.ndarray con: The computed connectivity matrix with a shape of
    (n_freqs, n_signals, n_signals).

    See Also
    --------
    For frequency-decomposition and frequency bin reference:
    https://docs.scipy.org/doc/numpy/reference/generated/numpy.fft.rfftfreq.html
    """
    # spacing between frequency bins
    spacing = epochs.info['sfreq'] / epochs.get_data().shape[-1]

    low_cutoff = tuple(band[0] for band in kwargs['freqs'])
    high_cutoff = tuple(
        band[1] - spacing if band != max(kwargs['freqs']) else band[1] for band
        in kwargs['freqs'])

    epochs_csd = compute_surface_laplacian(epochs=epochs, verbose=False)
    con, _, _, _, _ = spectral_connectivity(data=epochs_csd,
                                            method=kwargs['method'],
                                            sfreq=epochs.info['sfreq'],
                                            mode=kwargs['mode'],
                                            fmin=low_cutoff,
                                            fmax=high_cutoff,
                                            faverage=kwargs['faverage'],
                                            n_jobs=kwargs['n_jobs'],
                                            verbose=True)

    # from shape of (n_signals, n_signals, n_freqs) to
    # (n_freqs, n_signals, n_signals)
    con = np.transpose(con, (2, 0, 1))
    con = abs(con)

    return con


def compute_surface_laplacian(epochs: Epochs, verbose: bool = True) -> Epochs:
    """
    Performs the surface Laplacian transform on the Epochs instance
    For more information about the transform parameters please refer to
    Cohen, M. X. (2014). Analyzing neural time series data: theory and practice
    . MIT press. For more information about this function in particular,
    visit the MNE documentation at:
    https://mne.tools/dev/generated/mne.preprocessing.compute_current_source_density.html
    Parameters
    ----------
    epochs: the epochs to be transformed
    verbose: whether to visualize the power spectral densities before and after
    the Laplacian transform

    Returns
    -------
    Raw instance
    """

    epochs_csd = compute_current_source_density(epochs.copy())

    if verbose:
        fig, ax = plt.subplots(nrows=2,
                               ncols=1,
                               sharex='all',
                               sharey='all',
                               dpi=220)

        epochs.plot_psd(ax=ax[0], show=False, fmax=60)
        epochs_csd.plot_psd(ax=ax[1], show=False, fmax=60)
        plt.show()

    return epochs_csd



settings = dict(bandpass_filter=dict(low_freq=0.5, high_freq=45),
                epochs=dict(start_time=0., end_time=1.0, duration=1),
                ica=dict(n_components=32, method='infomax', decim=None),
                autoreject=dict(threshold=0.15),
                ransac=dict(threshold=0.2))
